--------------------------------------------------------------------------------

Satallax 3.4
============

M. Färber  
ENS Paris-Saclay, France


Introduction
------------

Satallax [Bro12] is an automated theorem prover for higher-order logic.
The particular form of higher-order logic supported by Satallax is
Church's simple type theory with extensionality and choice operators.
The SAT solver MiniSat [ES04] is responsible for much of the proof search.
The theoretical basis of search is a complete ground
tableau calculus for higher-order logic [BS10] with a choice operator [BB11].
Problems are given in the THF format.


Proof search
------------

A branch is formed from the axioms of the problem and
the negation of the conjecture (if any is given).
From this point on, Satallax tries to determine
unsatisfiability or satisfiability of this branch.
Satallax progressively generates higher-order formulae and
corresponding propositional clauses [Bro13].
These formulae and propositional clauses
correspond to instances of the tableau rules.
Satallax uses the SAT solver MiniSat to
test the current set of propositional clauses for unsatisfiability.
If the clauses are unsatisfiable, then the original branch is unsatisfiable.
Optionally, Satallax generates lambda-free higher-order logic (lfHOL) formulae
in addition to the propositional clauses [VBCS19].
If this option is used, then Satallax periodically calls the
theorem prover E [Sch13] to test for lfHOL unsatisfiability.
If the set of lfHOL formulae is unsatisfiable,
then the original branch is unsatisfiable.
Upon request, Satallax attempts to
reconstruct a proof which can be output in the TSTP format.


Strategies
----------

There are about 150 flags that control the order in which
formulae and instantiation terms are considered and
propositional clauses are generated.
Other flags activate some optional extensions to the basic proof procedure
(such as whether or not to call the theorem prover E).
A collection of flag settings is called a mode.
Approximately 500 modes have been defined and tested so far.
A strategy schedule is an ordered collection of modes with
information about how much time the mode should be allotted.
Satallax tries each of the modes for a certain amount of time sequentially.
Before deciding on the schedule to use,
Satallax parses the problem and determines if it is big enough that
a SInE-based premise selection algorithm [HV11] should be used.
If SInE is not activated, then Satallax uses
a strategy schedule consisting of 37 modes.
If SInE is activated, than Satallax is run with
a SInE-specific schedule consisting of
58 modes with different SInE parameter values selecting different premises.
Each mode is tried for time limits ranging
from less than a second to just over 1 minute.


Implementation
--------------

Satallax is implemented in OCaml, making use of the external tools
MiniSat (via a foreign function interface) and E.

Satallax is available at: <http://cl-informatik.uibk.ac.at/~mfaerber/satallax.html>


Expected Competition Performance
--------------------------------

Satallax 3.4 adds support for lambda-free higher order logic in E.
Among others, this allows Satallax to
efficiently translate formulas with partially applied terms to E.
This feature was implemented with the help of Petar Vukmirovic
during a research visit to Vrije Universiteit Amsterdam, which was
sponsored by a travel prize contributed by Jasmin Blanchette for CASC-J9.

This improves the performance of Satallax compared to version 3.3.
However, it might be offset by the rising amount of polymorphic problems in TPTP,
which Satallax does not support.


References
----------

Bro12
:   Brown, C.E. (2012),
    **Satallax: An Automated Higher-Order Prover (System Description)**,
    Gramlich, B. and Miller, D. and Sattler, U.,
    *Proceedings of the 6th International Joint Conference on Automated Reasoning*
    (Manchester, United Kingdom),
    pp. 111--117,
    Lecture Notes in Artificial Intelligence 7364,
    Springer-Verlag.

Bro13
:   Brown, C.E. (2013),
    **Reducing Higher-Order Theorem Proving to a Sequence of SAT Problems**,
    *Journal of Automated Reasoning* 51(1),
    pp. 57--77.

BB11
:   Backes, J. and Brown, C.E. (2011),
    **Analytic Tableaux for Higher-Order Logic with Choice**,
    *Journal of Automated Reasoning* 47(4),
    pp. 451--479.

BS10
:   Brown, C.E. and Smolka, G. (2010),
    **Analytic Tableaux for Simple Type Theory and its First-Order Fragment**,
    *Logical Methods in Computer Science* 6(2).

ES04
:   Eén, N. and Sörensson, N. (2004),
    **An Extensible SAT-solver**,
    Giunchiglia, E. and Tacchella, A.,
    *Proceedings of the 6th International Conference on Theory and Applications of Satisfiability Testing*
    (Santa Margherita Ligure, Italy),
    Lecture Notes in Computer Science 2919,
    pp. 502--518,
    Springer-Verlag.

HV11
:   Hoder, K. and Voronkov, A. (2011),
    **Sine Qua Non for Large Theory Reasoning**,
    Sofronie-Stokkermans, V. and Björner, N.,
    *Proceedings of the 23rd International Conference on Automated Deduction*
    (Wroclaw, Poland),
    pp. 299--314,
    Lecture Notes in Artificial Intelligence 6803,
    Springer-Verlag.

Sch13
:   Schulz, S. (2013)
    **System Description: E 1.8**,
    McMillan, K. and Middeldorp, A. and Voronkov, A.,
    *Proceedings of the 19th International Conference on Logic for Programming, Artificial Intelligence, and Reasoning*
    (Stellenbosch, South Africa),
    pp. 477--483,
    Lecture Notes in Computer Science 8312,
    Springer-Verlag.

VBCS19
:   Vukmirovic, P. and Blanchette, J.C. and Cruanes, S. and Schulz, S. (2019),
    **Extending a Brainiac Prover to Lambda-Free Higher-Order Logic**,
    Vojnar, T. and Zhang, L.,
    *Proceedings of the 25th International Conference on Tools and Algorithms for the Construction and Analysis of Systems*
    (Prague, Czech Republic),
    pp. 192--210,
    Lecture Notes in Computer Science 11427,
    Springer-Verlag.

--------------------------------------------------------------------------------
