open Flags

exception InputError of string

let comment_line_p l =
  if String.length l = 0 then true
  else l.[0] = ';'

let rec get_input_lines c =
  try 
    let line = input_line c in
    line :: get_input_lines c
  with End_of_file -> []

let get_content_lines c =
  List.filter (fun line -> not (comment_line_p line)) (get_input_lines c)

let rec set_flags = function
  | flag :: value :: rest -> set_flag flag value; set_flags rest
  | flag :: [] -> raise (InputError ("Value after flag " ^ flag ^ " expected"))
  | [] -> ()

let read_mode_file c = set_flags (get_content_lines c)

let read_schedule_line l =
  Scanf.sscanf l "%s %f" (fun mode time -> (mode, time))

let read_schedule_file c =
  List.map read_schedule_line (get_content_lines c)

let load_schedule s =
  let schedfile = !Config.modedir ^ "/" ^ s in
  if not (Sys.file_exists schedfile) then
    raise (InputError ("Could not find schedule " ^ schedfile));
  FileE.with_in schedfile read_schedule_file

let load_mode m =
  let modefile = (!Config.modedir ^ "/" ^ m) in
  if (not (Sys.file_exists modefile)) then
    raise (InputError ("Could not find mode " ^ modefile));
  FileE.with_in modefile read_mode_file

let parse_override s =
  match Str.split (Str.regexp "=") s with
  | [key;value] -> (key, value)
  | _ -> raise (InputError ("Could not parse flag/value pair " ^ s))

let parse_mode_overrides m =
  match Str.split (Str.regexp ":") m with
  | [] -> raise (InputError "Can not parse empty mode")
  | mode::overrides -> (mode, List.map parse_override overrides)

let load_mode_overrides (mode, overrides) =
  load_mode mode;
  List.iter (fun (k, v) -> set_flag k v) overrides
