open Problem
open Term
open TermP

let next_fresh_hyp : int ref = ref 0

(** Input: unit
	Output: returns next fresh hypothesis name **)
let rec get_hyp_name () =
	let x = "H" ^ (string_of_int (!next_fresh_hyp)) in
	incr next_fresh_hyp;
	if (Hashtbl.mem coq_used_names x) 
	then get_hyp_name ()
  	else x 

let rec find_fresh_consts n const =
  begin
    match n with 
    | Name(x,a) ->
	let x =try Hashtbl.find coq_names x with
	  Not_found
	  ->
	    failwith ("add_fresh_const can't find "^x^" in coq_names") in
	if List.mem_assoc x const then [] else [(x,a)] 
    | Ap(m1,m2) -> find_fresh_consts m1 const @ find_fresh_consts m2 const
    | Lam(_,m) -> find_fresh_consts m const
    | _ -> []
  end

(** [add_fresh_consts f m const] finds fresh constants in [m] wrt. [const],
    executes [f] on each of them and returns their union with [const].
    This is used for printing inhabitation tactics. *)
let add_fresh_consts f m const =
  let consts = find_fresh_consts (coqnorm m) const in
  let add_fresh_const cons (x, a) =
    if List.mem (x, a) cons then cons
    else (f x a; (x, a) :: cons) in
  List.fold_left add_fresh_const const consts
