open Args
open Atom
open Error
open Problem
open Refut
open Term

(*** Chad, August 2011: Added this way of printing proof information at the request of Jasmin Blanchette. ***)
let print_hocore c r =
  let don : (string,unit) Hashtbl.t = Hashtbl.create 1000 in
  let report_name name =
    if (not (Hashtbl.mem don name)) then
      begin
	Printf.fprintf c "%s\n" name;
	Hashtbl.add don name ()
      end
  in
  let rec hocore_1 c r a =
    if (!verbosity > 50) then
      begin
	Printf.printf "hocore_1 r\n";
	print_refut r;
	Printf.printf "hocore_1 a BEGIN\n";
	List.iter (fun (m,n) -> Printf.printf "%s : %s\n" n (trm_str m)) a;
	Printf.printf "hocore_1 a END\n";
      end;
    match r with
    | Eprover(pcore,cr,fcore,elines) ->
	begin
	  Printf.printf "Eprover\n propositional part:";
	  List.iter
	    (fun c ->
	      Printf.printf "%s\n" (String.concat " " (List.map string_of_int c));
	      List.iter
		(fun z -> (Printf.printf "Atom %d: %s\n" (abs z) (trm_str (Atom.atom_to_trm (abs z)))))
		c;
	      try Printf.printf "associated rule: %s\n" (ruleinfo_str (cr c))
	      with Not_found -> Printf.printf "no associated rule\n"
		  )
	    pcore;
	  Printf.printf "\n first-order part: ";
	  List.iter
	    (fun a -> Printf.printf " %d (%b)" a (try ignore(Atom.atom_to_trm (if a>0 then a else (-a))); true with Not_found -> false))
	    fcore;
	  Printf.printf "\n";
	end
    | SearchR(cl,cr) ->
	List.iter
	  (fun c0 ->
	    try
	      if (!verbosity > 50) then
		begin
		  Printf.printf "Calling cr on clause ";
		  List.iter (fun i -> Printf.printf "%d\n" i) c0;
		  Printf.printf "\n";
		end;
	      match (cr c0) with
	      |	ChoiceRule(Name(eps,_),p) ->
		  begin
		    if (!verbosity > 50) then Printf.printf "cr found it's a choice rule\n";
		    try
		      let (_,choiceax,_) = Hashtbl.find Choiceop.choiceopnames eps in
		      let name = List.assoc choiceax a in
		      report_name name
		    with
		      Not_found -> ()
		  end
	      |	ri ->
		  begin
		    if (!verbosity > 50) then Printf.printf "cr found it has rule info %s\n" (ruleinfo_str ri);
		  end
	    with (** Assumptions have no associated rule info **)
	      Not_found ->
		begin
		  if (!verbosity > 50) then Printf.printf "cr found no rule info\n";
		  match c0 with
		  | [l] ->
		      begin
			let tm = literal_to_trm l in
			try
			  report_name (List.assoc tm a)
			with Not_found -> ()
		      end
		  | _ -> raise (GenericError "Non-Unit Assumption") (*** Unexpected case: Non-unit Assumption ***)
		end)
	  cl
    | AssumptionConflictR(_) -> (*** There must be a formula (m,name) on a such that (~m,name2) is on a. Otherwise, there's a bug. ***)
	let rec find_conflict a =
	  match a with
	  | (((Ap(Ap(Imp,m),False)),name)::ar) ->
	      begin
		try
		  report_name (List.assoc m ar);
		  report_name name
		with
		| Not_found -> find_conflict ar
	      end
	  | ((m,name)::ar) ->
	      let nm = Ap(Ap(Imp,m),False) in
	      begin
		try
		  report_name (List.assoc nm ar);
		  report_name name
		with
		| Not_found -> find_conflict ar
	      end
	  | [] -> raise (GenericError "Assumption Conflict Not Found")
	in
	find_conflict a
    | FalseR ->
	begin
	  try
	    report_name (List.assoc False a)
	  with Not_found -> raise (GenericError "Assumption 'False' Not Found")
	end
    | NegReflR(_) ->
	let rec find_conflict a =
	  match a with
	  | (((Ap(Ap(Imp,Ap(Ap(Eq(_),m),n)),False)),name)::_) when (m = n) -> report_name name
	  | (_::ar) -> find_conflict ar
	  | [] -> raise (GenericError "Assumption of Negated Reflexive Equation Not Found")
	in
	find_conflict a
	  (*** Otherwise, follow the refutation, adding things to the assoc list a as we break down the formulas. ***)
    | NegImpR(m,n,r) ->
	begin
	  try
	    let name = List.assoc (neg (imp m n)) a in
	    hocore_1 c r ((m,name)::(normneg n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (imp m n)))))
	end
    | ImpR(m,n,r1,r2) ->
	begin
	  try
	    let name = List.assoc (imp m n) a in
	    hocore_1 c r1 ((normneg m,name)::a);
	    hocore_1 c r2 ((n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (imp m n))))
	end
    | NegAllR(b,m,x,r) -> 
	begin
	  try
	    let name = List.assoc (neg (Ap(Forall(b),m))) a in
	    hocore_1 c r (((norm name_def (neg (Ap(m,Name(x,b))))),name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (Ap(Forall(b),m))))))
	end
    | NegEqFR(a1,a2,m,n,r) -> 
	begin
	  try
	    let name = List.assoc (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))) a in
	    hocore_1 c r (((norm name_def (normneg (forall a1 (eq a2 (Ap(m,DB(0,a1))) (Ap(n,DB(0,a1))))))),name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))))))
	end
    | EqOR(m,n,r1,r2) -> 
	begin
	  try
	    let name = List.assoc (eq Prop m n) a in
	    hocore_1 c r1 ((m,name)::(normneg n,name)::a);
	    hocore_1 c r2 ((normneg m,name)::(n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (eq Prop m n))))
	end
    | NegEqOR(m,n,r1,r2) -> 
	begin
	  try
	    let name = List.assoc (neg (eq Prop m n)) a in
	    hocore_1 c r1 ((m,name)::(normneg n,name)::a);
	    hocore_1 c r2 ((normneg m,name)::(n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (eq Prop m n)))))
	end
  in
  let a = Hashtbl.fold (fun k v l -> (v,k)::l) name_hyp []
  in
  hocore_1 c r
    begin
      match !conjecture with
      | Some (_,ntmn) ->
	  if (!verbosity > 20) then Printf.printf "neg conj name %s %s\n" !conjecturename (trm_str ntmn);
	  ((ntmn,!conjecturename)::a)
      | _ -> a
    end
