open Args
open Atom
open Clauses
open Error
open Problem
open Refut
open Term

(*** Chad, August 2016 *)
let print_pfformdeps c r o =
  let handledclause : (int list,unit) Hashtbl.t = Hashtbl.create 1000 in
  let formdepsform : (int,bool) Hashtbl.t = Hashtbl.create 1000 in
  let formdepsfrom : (int,int list) Hashtbl.t = Hashtbl.create 1000 in
  let formdepsinsts : (int,stp * trm) Hashtbl.t = Hashtbl.create 1000 in
  let record_formdeps_common c0 b =
    match c0 with
    | (h::r) ->
	if not (Hashtbl.mem formdepsform (-h)) then Hashtbl.add formdepsform (-h) b;
	List.iter
	  (fun l ->
	    if not (Hashtbl.mem formdepsform l) then Hashtbl.add formdepsform l b;
	    Hashtbl.add formdepsfrom l [-h])
	  r
    | [] -> raise (Failure "Clause for common rule with no head?")
  in
  let record_formdeps_conn c0 b =
    match c0 with
    | (h1::h2::r) ->
	if not (Hashtbl.mem formdepsform (-h1)) then Hashtbl.add formdepsform (-h1) b;
	if not (Hashtbl.mem formdepsform (-h2)) then Hashtbl.add formdepsform (-h2) b;
	List.iter
	  (fun l ->
	    if not (Hashtbl.mem formdepsform l) then Hashtbl.add formdepsform l b;
	    Hashtbl.add formdepsfrom l [-h1;-h2])
	  r
    | [h1] -> raise (Failure "Clause for conn rule with only one head?")
    | [] -> raise (Failure "Clause for conn rule with no head?")
  in
  let rectopfrom l z =
    Hashtbl.add formdepsfrom z [l]
  in
  let handle_clause b cr c0 =
    if not (Hashtbl.mem handledclause c0) then
      try
	Hashtbl.add handledclause c0 ();
	match (cr c0) with
	| DeltaRule -> record_formdeps_common c0 b
	| NegPropRule(_) -> record_formdeps_common c0 b
	| PosPropRule(_) -> record_formdeps_common c0 b
	| InstRule(a,m,n) ->
	    record_formdeps_common c0 b;
	    begin
	      match c0 with
	      | (h::_) -> Hashtbl.add formdepsinsts (-h) (a,n)
	      | [] -> raise (Failure "InstRule with no head?")
	    end
	| FreshRule(_,_,_) -> record_formdeps_common c0 b
	| MatingRule(_,_) -> record_formdeps_conn c0 b
	| ConfrontationRule(_,_) -> record_formdeps_conn c0 b
	|	ChoiceRule(_,p) ->
	    begin
	      List.iter
		(fun l -> Hashtbl.add formdepsform l b)
		c0
	    end
	| Known(l,x,_) ->
	    Hashtbl.add formdepsform l b
      with (** Assumptions have no associated rule info **)
	Not_found ->
	  begin
	    match c0 with
	    | [l] ->
		Hashtbl.add formdepsform l b
	    | _ -> raise (GenericError "Non-Unit Assumption") (*** Unexpected case: Non-unit Assumption ***)
	  end
  in
  let rec pfformdeps_1 c r a =
    match r with
    | Eprover(pcore,cr,fcore,elines) -> List.iter (handle_clause true cr) pcore (*** questionable, but ignore fcore for now ***)
    | SearchR(cl,cr) -> List.iter (handle_clause true cr) cl
    | AssumptionConflictR(_) -> (*** There must be a formula (m,name) on a such that (~m,name2) is on a. Otherwise, there's a bug. ***)
	let find_conflict a =
	  match a with
	  | (((Ap(Ap(Imp,m),False)),name)::ar) ->
	      let l = get_literal m in
	      Hashtbl.add formdepsform l true;
	      Hashtbl.add formdepsform (-l) true
	  | ((m,name)::ar) ->
	      let nm = Ap(Ap(Imp,m),False) in
	      let l = get_literal nm in
	      Hashtbl.add formdepsform l true;
	      Hashtbl.add formdepsform (-l) true
	  | [] -> ()
	in
	find_conflict a
    | FalseR ->
	begin
	  try
	    Hashtbl.add formdepsform (get_literal False) true
	  with Not_found -> raise (GenericError "Assumption 'False' Not Found")
	end
    | NegReflR(_) ->
	let rec find_conflict a =
	  match a with
	  | ((((Ap(Ap(Imp,Ap(Ap(Eq(_),m),n)),False)) as emm),name)::_) when (m = n) ->
	      let lemm = get_literal emm in
	      Hashtbl.add formdepsform lemm true
	  | (_::ar) -> find_conflict ar
	  | [] -> raise (GenericError "Assumption of Negated Reflexive Equation Not Found")
	in
	find_conflict a
	  (*** Otherwise, follow the refutation, adding things to the assoc list a as we break down the formulas. ***)
    | NegImpR(m,n,r) ->
	begin
	  try
	    let mn = neg (imp m n) in
	    let name = List.assoc mn a in
	    let lmn = get_literal mn in
	    Hashtbl.add formdepsform lmn true;
	    rectopfrom lmn (get_literal m);
	    rectopfrom lmn (-(get_literal n));
	    pfformdeps_1 c r ((m,name)::(normneg n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (imp m n)))))
	end
    | ImpR(m,n,r1,r2) ->
	begin
	  try
	    let mn = imp m n in
	    let name = List.assoc mn a in
	    let lmn = get_literal mn in
	    Hashtbl.add formdepsform lmn true;
	    rectopfrom lmn (-(get_literal m));
	    rectopfrom lmn (get_literal n);
	    pfformdeps_1 c r1 ((normneg m,name)::a);
	    pfformdeps_1 c r2 ((n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (imp m n))))
	end
    | NegAllR(b,m,x,r) -> 
	begin
	  try
	    let name = List.assoc (neg (Ap(Forall(b),m))) a in
	    let l1 = get_literal (neg (Ap(Forall(b),m))) in
	    let l2 = get_literal (norm name_def (neg (Ap(m,Name(x,b))))) in
	    Hashtbl.add formdepsform l1 true;
	    rectopfrom l1 l2;
	    pfformdeps_1 c r (((norm name_def (neg (Ap(m,Name(x,b))))),name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (Ap(Forall(b),m))))))
	end
    | NegEqFR(a1,a2,m,n,r) -> 
	begin
	  try
	    let name = List.assoc (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))) a in
	    let l1 = get_literal (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))) in
	    let l2 = get_literal (norm name_def (normneg (forall a1 (eq a2 (Ap(m,DB(0,a1))) (Ap(n,DB(0,a1))))))) in
	    Hashtbl.add formdepsform l1 true;
	    rectopfrom l1 l2;
	    pfformdeps_1 c r (((norm name_def (normneg (forall a1 (eq a2 (Ap(m,DB(0,a1))) (Ap(n,DB(0,a1))))))),name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))))))
	end
    | EqOR(m,n,r1,r2) -> 
	begin
	  try
	    let name = List.assoc (eq Prop m n) a in
	    let l1 = get_literal (eq Prop m n) in
	    let lm = get_literal m in
	    let ln = get_literal n in
	    Hashtbl.add formdepsform l1 true;
	    rectopfrom l1 lm;
	    rectopfrom l1 ln;
	    rectopfrom l1 (-lm);
	    rectopfrom l1 (-ln);
	    pfformdeps_1 c r1 ((m,name)::(normneg n,name)::a);
	    pfformdeps_1 c r2 ((normneg m,name)::(n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (eq Prop m n))))
	end
    | NegEqOR(m,n,r1,r2) -> 
	begin
	  try
	    let name = List.assoc (neg (eq Prop m n)) a in
	    let l1 = -(get_literal (eq Prop m n)) in
	    let lm = get_literal m in
	    let ln = get_literal n in
	    Hashtbl.add formdepsform l1 true;
	    rectopfrom l1 lm;
	    rectopfrom l1 ln;
	    rectopfrom l1 (-lm);
	    rectopfrom l1 (-ln);
	    pfformdeps_1 c r1 ((m,name)::(normneg n,name)::a);
	    pfformdeps_1 c r2 ((normneg m,name)::(n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (eq Prop m n)))))
	end
  in
  let a = Hashtbl.fold (fun k v l -> (v,k)::l) name_hyp []
  in
  pfformdeps_1 c r
    begin
      match !conjecture with
      | Some (_,ntmn) ->
	  if (!verbosity > 20) then Printf.printf "neg conj name %s %s\n" !conjecturename (trm_str ntmn);
	  ((ntmn,!conjecturename)::a)
      | _ -> a
    end;
  (***
  Printf.printf "allclauses:\n";
  List.iter (fun c0 ->
    List.iter (fun l -> Printf.printf "%d " l) c0;
    if Hashtbl.mem handledclause c0 then
      Printf.printf "(used)\n"
    else
      Printf.printf "\n")
    !allclauses;
     ***)
  List.iter
    (handle_clause false (Hashtbl.find !allclause_ruleinfo))
    !allclauses;
  let usefulcnt = ref 0 in
  let uselesscnt = ref 0 in
  let don : (int,bool * trm) Hashtbl.t = Hashtbl.create 1000 in
  let depslisth : (int * int list,unit) Hashtbl.t = Hashtbl.create 1000 in
  let depslist : (int * int list) list ref = ref [] in
  Hashtbl.iter
    (fun k v ->
      if not (Hashtbl.mem don k) then
	begin
	  if v then incr usefulcnt else incr uselesscnt;
	  Hashtbl.add don k (v,literal_to_trm k)
	end)
    formdepsform;
  Printf.printf "%d formulas generated, %d useful, %d useless\n" (!usefulcnt + !uselesscnt) !usefulcnt !uselesscnt;
  Hashtbl.iter
    (fun k _ ->
      let ll = Hashtbl.find_all formdepsfrom k in
      List.iter (fun v -> if not (Hashtbl.mem depslisth (k,v)) then (depslist := (k,v)::!depslist; Hashtbl.add depslisth (k,v) ())) ll)
   don;
  match o with
  | Some(o) ->
      let f = open_out_bin o in
      output_value f don;
      output_value f !depslist;
      close_out f
  | None -> ()
