open Args
open Atom
open Error
open Problem
open Refut
open Term

(*** Chad, August 2016 *)
let print_pfuseful c r o =
  let usefulform : (int,string) Hashtbl.t = Hashtbl.create 1000 in
  let usefulfrom : (int,int) Hashtbl.t = Hashtbl.create 1000 in
  let usefulinsts : (int,stp * trm) Hashtbl.t = Hashtbl.create 1000 in
  let record_useful_common c0 =
    match c0 with
    | (h::r) ->
	if not (Hashtbl.mem usefulform (-h)) then Hashtbl.add usefulform (-h) "";
	List.iter
	  (fun l ->
	    if not (Hashtbl.mem usefulform l) then Hashtbl.add usefulform l "";
	    Hashtbl.add usefulfrom (-h) l)
	  r
    | [] -> raise (Failure "Clause for common rule with no head?")
  in
  let record_useful_conn c0 =
    match c0 with
    | (h1::h2::r) ->
	if not (Hashtbl.mem usefulform (-h1)) then Hashtbl.add usefulform (-h1) "";
	if not (Hashtbl.mem usefulform (-h2)) then Hashtbl.add usefulform (-h2) "";
	List.iter
	  (fun l ->
	    if not (Hashtbl.mem usefulform l) then Hashtbl.add usefulform l "";
	    Hashtbl.add usefulfrom (-h1) l;
	    Hashtbl.add usefulfrom (-h2) l)
	  r
    | [h1] -> raise (Failure "Clause for conn rule with only one head?")
    | [] -> raise (Failure "Clause for conn rule with no head?")
  in
  let rectopfrom l z =
    Hashtbl.add usefulfrom l z
  in
  let record_useful_clause_ruleinfo a cr c0 =
	    try
	      match (cr c0) with
	      | DeltaRule -> record_useful_common c0
	      | NegPropRule(_) -> record_useful_common c0
	      | PosPropRule(_) -> record_useful_common c0
	      | InstRule(a,m,n) ->
		  record_useful_common c0;
		  begin
		    match c0 with
		    | (h::_) -> Hashtbl.add usefulinsts (-h) (a,n)
		    | [] -> raise (Failure "InstRule with no head?")
		  end
	      | FreshRule(_,_,_) -> record_useful_common c0
	      | MatingRule(_,_) -> record_useful_conn c0
	      | ConfrontationRule(_,_) -> record_useful_conn c0
	      |	ChoiceRule(Name(eps,_),p) ->
		  let choicename =
		    try
		      let (_,choiceax,_) = Hashtbl.find Choiceop.choiceopnames eps in
		      let name = List.assoc choiceax a in
		      ("choice:" ^ name)
		    with
		      Not_found -> "choice"
		  in
		  begin
		    List.iter
		      (fun l -> Hashtbl.add usefulform l choicename)
		      c0
		  end
	      |	ChoiceRule(_,p) ->
		  begin
		    List.iter
		      (fun l -> Hashtbl.add usefulform l "choice")
		      c0
		  end
	      | Known(l,x,_) ->
		  Hashtbl.add usefulform l ("known:" ^ itp_thm_str x)
	    with (** Assumptions have no associated rule info **)
	      Not_found ->
		begin
		  match c0 with
		  | [l] ->
		      let name =
			try
			  "assumption:" ^ (List.assoc (literal_to_trm l) a)
			with
			  Not_found -> "assumption"
		      in		  
		      Hashtbl.add usefulform l name
		  | _ -> raise (GenericError "Non-Unit Assumption") (*** Unexpected case: Non-unit Assumption ***)
		end
  in
  let rec pfuseful_1 c r a =
    match r with
    | Eprover(pcore,cr,fcore,elines) -> (*** questionable, but ignore fcore for now ***)
	List.iter (record_useful_clause_ruleinfo a cr) pcore
    | SearchR(cl,cr) ->
	List.iter (record_useful_clause_ruleinfo a cr) cl
    | AssumptionConflictR(_) -> (*** There must be a formula (m,name) on a such that (~m,name2) is on a. Otherwise, there's a bug. ***)
	let rec find_conflict a =
	  match a with
	  | (((Ap(Ap(Imp,m),False)),name)::ar) ->
	      begin
		try
		  let name2 = List.assoc m ar in
		  let l = get_literal m in
		  Hashtbl.add usefulform l ("top:" ^ name);
		  Hashtbl.add usefulform (-l) ("top:" ^ name2);
		with
		| Not_found -> find_conflict ar
	      end
	  | ((m,name)::ar) ->
	      let nm = Ap(Ap(Imp,m),False) in
	      begin
		try
		  let name2 = List.assoc nm ar in
		  let l = get_literal nm in
		  Hashtbl.add usefulform l ("top:" ^ name);
		  Hashtbl.add usefulform (-l) ("top:" ^ name2)
		with
		| Not_found -> find_conflict ar
	      end
	  | [] -> raise (GenericError "Assumption Conflict Not Found")
	in
	find_conflict a
    | FalseR ->
	begin
	  try
	    let name = List.assoc False a in
	    Hashtbl.add usefulform (get_literal False) ("top:" ^ name)
	  with Not_found -> raise (GenericError "Assumption 'False' Not Found")
	end
    | NegReflR(_) ->
	let rec find_conflict a =
	  match a with
	  | ((((Ap(Ap(Imp,Ap(Ap(Eq(_),m),n)),False)) as emm),name)::_) when (m = n) ->
	      let lemm = get_literal emm in
	      Hashtbl.add usefulform lemm ("top:" ^ name)
	  | (_::ar) -> find_conflict ar
	  | [] -> raise (GenericError "Assumption of Negated Reflexive Equation Not Found")
	in
	find_conflict a
	  (*** Otherwise, follow the refutation, adding things to the assoc list a as we break down the formulas. ***)
    | NegImpR(m,n,r) ->
	begin
	  try
	    let mn = neg (imp m n) in
	    let name = List.assoc mn a in
	    let lmn = get_literal mn in
	    Hashtbl.add usefulform lmn ("top:" ^ name);
	    rectopfrom lmn (get_literal m);
	    rectopfrom lmn (-(get_literal n));
	    pfuseful_1 c r ((m,name)::(normneg n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (imp m n)))))
	end
    | ImpR(m,n,r1,r2) ->
	begin
	  try
	    let mn = imp m n in
	    let name = List.assoc mn a in
	    let lmn = get_literal mn in
	    Hashtbl.add usefulform lmn ("top:" ^ name);
	    rectopfrom lmn (-(get_literal m));
	    rectopfrom lmn (get_literal n);
	    pfuseful_1 c r1 ((normneg m,name)::a);
	    pfuseful_1 c r2 ((n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (imp m n))))
	end
    | NegAllR(b,m,x,r) -> 
	begin
	  try
	    let name = List.assoc (neg (Ap(Forall(b),m))) a in
	    let l1 = get_literal (neg (Ap(Forall(b),m))) in
	    let l2 = get_literal (norm name_def (neg (Ap(m,Name(x,b))))) in
	    Hashtbl.add usefulform l1 ("top:" ^ name);
	    rectopfrom l1 l2;
	    pfuseful_1 c r (((norm name_def (neg (Ap(m,Name(x,b))))),name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (Ap(Forall(b),m))))))
	end
    | NegEqFR(a1,a2,m,n,r) -> 
	begin
	  try
	    let name = List.assoc (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))) a in
	    let l1 = get_literal (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))) in
	    let l2 = get_literal (norm name_def (normneg (forall a1 (eq a2 (Ap(m,DB(0,a1))) (Ap(n,DB(0,a1))))))) in
	    Hashtbl.add usefulform l1 ("top:" ^ name);
	    rectopfrom l1 l2;
	    pfuseful_1 c r (((norm name_def (normneg (forall a1 (eq a2 (Ap(m,DB(0,a1))) (Ap(n,DB(0,a1))))))),name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (Ap(Ap(Eq(Ar(a1,a2)),m),n))))))
	end
    | EqOR(m,n,r1,r2) -> 
	begin
	  try
	    let name = List.assoc (eq Prop m n) a in
	    let l1 = get_literal (eq Prop m n) in
	    let lm = get_literal m in
	    let ln = get_literal n in
	    Hashtbl.add usefulform l1 ("top:" ^ name);
	    rectopfrom l1 lm;
	    rectopfrom l1 ln;
	    rectopfrom l1 (-lm);
	    rectopfrom l1 (-ln);
	    pfuseful_1 c r1 ((m,name)::(normneg n,name)::a);
	    pfuseful_1 c r2 ((normneg m,name)::(n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (eq Prop m n))))
	end
    | NegEqOR(m,n,r1,r2) -> 
	begin
	  try
	    let name = List.assoc (neg (eq Prop m n)) a in
	    let l1 = -(get_literal (eq Prop m n)) in
	    let lm = get_literal m in
	    let ln = get_literal n in
	    Hashtbl.add usefulform l1 ("top:" ^ name);
	    rectopfrom l1 lm;
	    rectopfrom l1 ln;
	    rectopfrom l1 (-lm);
	    rectopfrom l1 (-ln);
	    pfuseful_1 c r1 ((m,name)::(normneg n,name)::a);
	    pfuseful_1 c r2 ((normneg m,name)::(n,name)::a)
	  with Not_found -> raise (GenericError ("Unknown Assumption " ^ (trm_str (neg (eq Prop m n)))))
	end
  in
  let a = Hashtbl.fold (fun k v l -> (v,k)::l) name_hyp []
  in
  pfuseful_1 c r
    begin
      match !conjecture with
      | Some (_,ntmn) ->
	  if (!verbosity > 20) then Printf.printf "neg conj name %s %s\n" !conjecturename (trm_str ntmn);
	  ((ntmn,!conjecturename)::a)
      | _ -> a
    end;
  Printf.printf "Useful formulas:\n";
  let don : (int,trm) Hashtbl.t = Hashtbl.create 1000 in
  let depslist : (int * int) list ref = ref [] in
  Hashtbl.iter
    (fun k _ ->
      if not (Hashtbl.mem don k) then
	begin
	  Printf.printf "%d %s\n" k (trm_str (literal_to_trm k));
	  List.iter (fun v -> if not (v = "") then Printf.printf "%s;" v) (Hashtbl.find_all usefulform k);
	  Printf.printf "\n";
	  Hashtbl.add don k (literal_to_trm k)
	end)
    usefulform;
  Printf.printf "Dependencies:\n";
  Hashtbl.iter
    (fun k _ ->
      let ll = Hashtbl.find_all usefulfrom k in
      if not (ll = []) then
	begin
	  Printf.printf "%d:" k;
	  List.iter (fun v -> if Hashtbl.mem usefulform v then (Printf.printf " %d" v; depslist := (v,k)::!depslist)) ll;
	  Printf.printf "\n";
	end)
    don;
  Printf.printf "Instantiations:\n";
  Hashtbl.iter
    (fun k _ ->
      let ll = Hashtbl.find_all usefulinsts k in
      if not (ll = []) then
	begin
	  Printf.printf "%d: (%d)\n" k (List.length ll);
	  List.iter (fun (_,n) -> Printf.printf "%s\n" (trm_str n)) ll
	end)
    don;
  match o with
  | Some(o) ->
      let f = open_out_bin o in
      output_value f don;
      output_value f !depslist;
      close_out f
  | None -> ()
