(*** Chad - July 2012 - output proof script in tstp format. ***)
(*** Chad, Dec 2016: Newer tstp proof: print defns of propositions referenced, defns of eigenvars using choice, propositional clauses using symbols for propositions corresponding to assumptions and rules, and a final step of false by propositional unsatisfiability **)

open Clauses
open Fresh
open Problem
open Refut
open Refutation
open Term
open TermP

(* copied from coq.ml *)

let forallbvarnames : (string,string) Hashtbl.t = Hashtbl.create 100;;
let nexistsbvarnames : (string,string) Hashtbl.t = Hashtbl.create 100;;


(*** July 2012 (Chad) : TSTP ***)
(** Prints type m as a Tstp-formatted string on the out_channel c  **)
let rec print_stp_tstp c m p =
  match m with
  | Base x ->
      Printf.fprintf c "%s" x
  | Prop ->
      Printf.fprintf c "$o"
  | Ar(a,b) ->
      begin
	if p then Printf.fprintf c "(";
	print_stp_tstp c a true;
	Printf.fprintf c ">";
	print_stp_tstp c b false;
	if p then Printf.fprintf c ")";
	flush c
      end

let add_fresh_const c const n sp =
  let add x a =
    Printf.fprintf c "%stab_inh (" sp;
    print_stp_tstp c a false;
    Printf.fprintf c ") %s.\n" x;
  in add_fresh_consts add n const

let tstp_defprops : (trm,string) Hashtbl.t = Hashtbl.create 100;;
let next_defprop_num = ref 1;;

let rec next_defprop_name used =
  let x = "sP" ^ (string_of_int !next_defprop_num) in
  incr next_defprop_num;
  if List.mem x used then
    next_defprop_name used
  else
    x

(** Input: out_channel c, term m, list of bound variables 
	Invariant: m is closed, if  it is enclosed in quantifiers for the bound variables 
	Prints the term m on the channel c**)
let rec trm_to_tstp c m bound =
  try
    let p = Hashtbl.find tstp_defprops m in
    Printf.fprintf c "%s" p
  with Not_found ->
    match m with
      Name(x,_) -> (* Definitions *)
	Printf.fprintf c "%s" (tstpizename x)
    | False -> (* Bottom *)
	Printf.fprintf c "$false"
    | Ap(Ap(Imp,m1),False) ->  (* Negation *)
	begin
	  Printf.fprintf c "(~(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c "))";
	end
    | Ap(Ap(Imp,m1),m2) -> (* Implication *)
	begin
	  Printf.fprintf c "(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c " => ";
	  trm_to_tstp c m2 bound;
	  Printf.fprintf c ")";
	end
    | Ap(Imp,m1) -> trm_to_tstp c (Lam(Prop,Ap(Ap(Imp,shift m1 0 1),DB(0,Prop)))) bound
    | Imp -> trm_to_tstp c (Lam(Prop,Lam(Prop,Ap(Ap(Imp,DB(1,Prop)),DB(0,Prop))))) bound
    | Ap(Forall(a),Lam(_,m1)) -> (* forall with Lam *)
	begin
	  print_all_tstp c a m1 bound
	end
    | Forall(a) ->
	begin
	  Printf.fprintf c "(!!)";
	end
    | Ap(Ap(Eq(a),m1),m2) -> (* Equality *)
	begin
	  Printf.fprintf c "(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c " = ";
	  trm_to_tstp c m2 bound;
	  Printf.fprintf c ")"
	end
    | Eq(a) ->     
	Printf.fprintf c "(=)"
    | Ap(Choice(a),Lam(_,m1)) ->
	let bound = Variables.push bound in
	Printf.fprintf c "(@+["; Printf.fprintf c "%s" (Variables.top bound); 
	Printf.fprintf c ":"; print_stp_tstp c a false; 
	Printf.fprintf c "]:";
	trm_to_tstp c m1 bound; Printf.fprintf c ")"
    | Ap(Choice(a),m1) ->
	trm_to_tstp c (Ap(Choice(a), Lam(a, Ap(shift m1 0 1, DB(0, a))))) bound;
    | Choice(a) ->
	(* Not valid TSTP, should not occur on its own *)
	Printf.fprintf c "(@+)";
    | True -> (* Top *)
	Printf.fprintf c "$true"
    | Ap(Ap(And,m1),m2) -> (* conjunction *)
	begin
	  Printf.fprintf c "(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c " & ";
	  trm_to_tstp c m2 bound;
	  Printf.fprintf c ")";
	end
    | And ->Printf.fprintf c "(&)"
    | Ap(Ap(Or,m1),m2) -> (* disjunction *)
	begin
	  Printf.fprintf c "(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c " | ";
	  trm_to_tstp c m2 bound;
	  Printf.fprintf c ")";
	end
    | Or -> Printf.fprintf c "(|)"
    | Ap(Ap(Iff,m1),m2) -> (* equivalenz *)
	begin
	  Printf.fprintf c "(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c " <=> ";
	  trm_to_tstp c m2 bound;
	  Printf.fprintf c ")";
	end
    | Iff -> Printf.fprintf c "(<=>)"
    | Neg -> Printf.fprintf c "(~)"
    | Ap(Exists(a),Lam(_,m1)) -> (* exist *)
	begin
	  print_ex_tstp c a m1 bound
	end
    | Exists(a) ->
	begin
	  Printf.fprintf c "(??)";
	end
    | DB(i,a) -> (* Bound variable *)
	Printf.fprintf c "%s" (Variables.get i bound)
    | Lam(a,m) ->
	begin
	  print_lam_tstp c a m bound
	end
    | Ap(m1,m2) ->     
	begin
	  Printf.fprintf c "(";
	  trm_to_tstp c m1 bound;
	  Printf.fprintf c " @ ";
	  trm_to_tstp c m2 bound;
	  Printf.fprintf c ")";
	end      

 (* Prints consecutive lambda-terms as a single fun in Tstp. *) 
and print_lam_tstp c a m bound =
	let bound = Variables.push bound in
	Printf.fprintf c "(^["; Printf.fprintf c "%s" (Variables.top bound); Printf.fprintf c ":"; print_stp_tstp c a false; Printf.fprintf c "]:";
	match m with
		| Lam(b,m') -> print_lam_tstp c b m' bound; Printf.fprintf c ")"
		| _ -> trm_to_tstp c m bound; Printf.fprintf c ")"

(* Prints consecutive forall-terms together with the corresponding lambda-terms as a single forall in Tstp. *) 		
and print_all_tstp c a m bound =
  let bound = Variables.push bound in
  Printf.fprintf c "(!["; Printf.fprintf c "%s" (Variables.top bound); Printf.fprintf c ":"; print_stp_tstp c a false; Printf.fprintf c "]:";
  match m with
  | Ap(Forall(a'),Lam(_,m')) -> print_all_tstp c a' m' bound; Printf.fprintf c ")"
  | _ -> trm_to_tstp c m bound; Printf.fprintf c ")"

(* Prints an exist-term together with the corresponding lambda-term as an exists in Tstp. *) 		
and print_ex_tstp c a m bound =
 	let bound = Variables.push bound in
	Printf.fprintf c "(?["; Printf.fprintf c "%s" (Variables.top bound); 
	Printf.fprintf c ":"; print_stp_tstp c a false; 
        Printf.fprintf c "]:";
	trm_to_tstp c m bound; Printf.fprintf c ")"

let tstp_axioms : (trm,string * trm * out_channel * (bool ref)) Hashtbl.t = Hashtbl.create 100;;

let trm_to_tstp_rembvar x c m bound =
  try
    let p = Hashtbl.find tstp_defprops m in
    Printf.fprintf c "%s" p
  with Not_found ->
    match m with
    | Ap(Forall(a),Lam(_,m1)) ->
	let bound = Variables.push bound in
	let y = Variables.top bound in
	Hashtbl.add forallbvarnames x y;
	Printf.fprintf c "(!["; Printf.fprintf c "%s" y; Printf.fprintf c ":"; print_stp_tstp c a false; Printf.fprintf c "]:";
	trm_to_tstp c m1 bound; Printf.fprintf c ")"
    | Ap(Neg,Ap(Exists(a),Lam(_,m1))) ->
	let bound = Variables.push bound in
	let y = Variables.top bound in
	Hashtbl.add nexistsbvarnames x y;
	Printf.fprintf c "(~(?["; Printf.fprintf c "%s" y; Printf.fprintf c ":"; print_stp_tstp c a false; Printf.fprintf c "]:";
	trm_to_tstp c m1 bound; Printf.fprintf c "))"
    | Ap(Ap(Imp,Ap(Exists(a),Lam(_,m1))),False) ->
	let bound = Variables.push bound in
	let y = Variables.top bound in
	Hashtbl.add nexistsbvarnames x y;
	Printf.fprintf c "(~(?["; Printf.fprintf c "%s" y; Printf.fprintf c ":"; print_stp_tstp c a false; Printf.fprintf c "]:";
	trm_to_tstp c m1 bound; Printf.fprintf c "))"
    | _ ->
	trm_to_tstp c m bound

let lookup_tstp w s hyp =
  try
    List.assoc s hyp
  with
  | Not_found ->
      begin
	try
	  let (x,m,c,f) = Hashtbl.find tstp_axioms s in
	  if (!f) then
	    begin
	      f := false;
	      Printf.fprintf c "thf(%s,axiom," x;
	      trm_to_tstp_rembvar x c m (Variables.make ());
	      Printf.fprintf c ").\n";
	      flush c
	    end;
	  x
	with
	| Not_found ->
	  Printf.printf "%s: Could not find hyp name\ns = %s\nhyp:\n" w (trm_str s);
	  List.iter (fun (m,h) -> Printf.printf "%s: %s\n" h (trm_str m)) hyp;
	  failwith ("Could not find hyp name")
      end

let tstp_print_defprop c used s =
  let x = next_defprop_name used in
  Printf.fprintf c "thf(%s,plain,%s <=> " x x;
  trm_to_tstp_rembvar x c s (Variables.make ());
  Printf.fprintf c ",introduced(definition,[new_symbols(definition,[%s])])).\n" x;
  Hashtbl.add tstp_defprops s x

let tstp_print_eigendef c tstp_name def =
  output_string c ("thf(eigendef_" ^ tstp_name ^ ", definition, " ^ tstp_name ^ " = ");
  trm_to_tstp c def (Variables.make());
  output_string c (", introduced(definition,[new_symbols(definition,[" ^ tstp_name ^ "])])).\n")


let tstpline : int ref = ref 1;;

let tstphyp : (trm,string) Hashtbl.t = Hashtbl.create 100

(**
 Input: unit
	Output: returns next fresh hypothesis name **)
let rec get_thyp_name () =
  let x = "h" ^ (string_of_int (!next_fresh_hyp)) in
  incr next_fresh_hyp;
  if (Hashtbl.mem coq_used_names x) 
  then get_thyp_name ()
  else
    x

let rec get_tstp_hyp_name c s =
  try
    Hashtbl.find tstphyp s
  with Not_found ->
    let x = "h" ^ (string_of_int (!next_fresh_hyp)) in
    Hashtbl.add tstphyp s x;
    incr next_fresh_hyp;
    if (Hashtbl.mem coq_used_names x) 
    then get_tstp_hyp_name c s
    else
      begin
	Printf.fprintf c "thf(%s,assumption," x;
	trm_to_tstp_rembvar x c s (Variables.make ());
	Printf.fprintf c ",introduced(assumption,[])).\n";
	flush c;
	x
      end

let rec as_str2 al =
  match al with
  | (a::ar) -> "," ^ a ^ (as_str2 ar)
  | [] -> ""

let as_str al =
  match al with
  | (a::ar) -> "[" ^ a ^ (as_str2 ar) ^ "]"
  | [] -> "[]"

let rec disch_str2 d =
  match d with
  | (x::r) -> "," ^ x ^ (disch_str2 r)
  | [] -> ""

let disch_str d =
  match d with
  | (x::r) -> x ^ (disch_str2 r)
  | [] -> ""

let rec info_str2 r dl =
  match dl with
  | (d::dr) -> "," ^ r ^ "(discharge,[" ^ (disch_str d) ^ "])" ^ (info_str2 r dr)
  | [] -> ""

let info_str r hyp dl =
  match hyp with
  | [] -> info_str2 r dl
  | _ -> ",assumptions(" ^ (as_str (List.map (fun (_,h) -> h) hyp)) ^ ")" ^ (info_str2 r dl)

(** Input: out_channel c, refutation r, association list (term -> hypothesis name) hyp, association list (constant name -> type) const, Space string sp 
	Output: unit, prints refutation r to c **)
let rec ref_tstp1 c r hyp const =
  let l = !tstpline in
  incr tstpline;
  begin
    match r with
    | Conflict(s,ns) ->
	let p1 = (lookup_tstp "0" (coqnorm s) hyp) in
	let p2 = (lookup_tstp "1" (coqnorm ns) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_conflict,[status(thm)%s],[%s,%s])).\n" l (info_str "tab_conflict" hyp []) p1 p2
    | Fal(_) -> 				
	let p1 = (lookup_tstp "2" False hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_false,[status(thm)%s],[%s])).\n" l (info_str "tab_false" hyp []) p1
    | NegRefl(s) -> 			
	let p1 = (lookup_tstp "3" (coqnorm s) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_refl,[status(thm)%s],[%s])).\n" l (info_str "tab_refl" hyp []) p1;
    | Implication(h,s,t,r1,r2) -> 	
	let h1 = get_tstp_hyp_name c (coqnorm s) in	
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm t,h2)::hyp) const in
	let p1 = (lookup_tstp "4" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_imp,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_imp" hyp [[h1];[h2]]) p1 l1 l2 (disch_str2 [h1;h2]);
    | Disjunction(h,s,t,r1,r2) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in	
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm t,h2)::hyp) const in
	let p1 = (lookup_tstp "5" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_or,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_or" hyp [[h1];[h2]]) p1 l1 l2 (disch_str2 [h1;h2]);
    | NegConjunction(h,s,t,r1,r2) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in	
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm t,h2)::hyp) const in
	let p1 = (lookup_tstp "6" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_nand,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_nand" hyp [[h1];[h2]]) p1 l1 l2 (disch_str2 [h1;h2]);
    | NegImplication(h,s,t,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm t,h2)::hyp) const in
	let p1 = (lookup_tstp "7" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_negimp,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_negimp" hyp [[h1;h2]]) p1 l1 (disch_str2 [h1;h2]);
    | Conjunction(h,s,t,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm t,h2)::hyp) const in
	let p1 = (lookup_tstp "8" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_and,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_and" hyp [[h1;h2]]) p1 l1 (disch_str2 [h1;h2]);
    | NegDisjunction(h,s,t,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm t,h2)::hyp) const in
	let p1 = (lookup_tstp "9" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_nor,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_nor" hyp [[h1;h2]]) p1 l1 (disch_str2 [h1;h2]);
    | All(h,s,r1,a,m,n) ->
	let const = add_fresh_const c const n "" in
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let p1 = (lookup_tstp "9" (coqnorm h) hyp) in
	begin
	  try
	    let x = Hashtbl.find forallbvarnames p1 in
	    Printf.fprintf c "thf(%d,plain,$false,inference(tab_all,[status(thm)%s],[%s:[bind(%s,$thf(" l (info_str "tab_all" hyp [[h1]]) p1 x;
	    trm_to_tstp c n (Variables.make ());
	    Printf.fprintf c "))],%d%s])).\n" l1 (disch_str2 [h1])
	  with Not_found ->
	    Printf.fprintf c "thf(%d,plain,$false,inference(tab_all,[status(thm)%s],[%s:[bind(Xnoname,$thf(" l (info_str "tab_all" hyp [[h1]]) p1;
	    trm_to_tstp c n (Variables.make ());
	    Printf.fprintf c "))],%d%s])).\n" l1 (disch_str2 [h1])
	end
    | NegAll(h,s,r1,a,m,x) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) ((x,a)::const) in
	let p1 = (lookup_tstp "11" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_negall,[status(thm)%s,tab_negall(eigenvar,%s)],[%s,%d%s])).\n" l (info_str "tab_negall" hyp [[h1]]) (tstpizename x) p1 l1 (disch_str2 [h1])
    | Exist(h,s,r1,a,m,x) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) ((x,a)::const) in
	let p1 = (lookup_tstp "11" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_ex,[status(thm)%s,tab_ex(eigenvar,%s)],[%s,%d%s])).\n" l (info_str "tab_ex" hyp [[h1]]) (tstpizename x) p1 l1 (disch_str2 [h1])
    | NegExist(h,s,r1,a,m,n) ->
	let const = add_fresh_const c const n "" in
	let h1 = get_tstp_hyp_name c (coqnorm s) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let p1 = (lookup_tstp "12" (coqnorm h) hyp) in
	begin
	  try
	    let x = Hashtbl.find nexistsbvarnames p1 in
	    Printf.fprintf c "thf(%d,plain,$false,inference(tab_negex,[status(thm)%s],[%s:[bind(%s,$thf(" l (info_str "tab_negex" hyp [[h1]]) p1 x;
	    trm_to_tstp c n (Variables.make ());
	    Printf.fprintf c "))],%d%s])).\n" l1 (disch_str2 [h1])
	  with Not_found ->
	    Printf.fprintf c "thf(%d,plain,$false,inference(tab_negex,[status(thm)%s],[%s:[bind(Xnoname,$thf(" l (info_str "tab_negex" hyp [[h1]]) p1;
	    trm_to_tstp c n (Variables.make ());
	    Printf.fprintf c "))],%d%s])).\n" l1 (disch_str2 [h1])
	end
    | Mating(h1,h2, ss, rs) ->
	let hl = ref [] in
	let ll = ref [] in
	List.iter (fun (s,r) ->
	  let h' = get_tstp_hyp_name c (coqnorm s) in
	  let l' = ref_tstp1 c r ((coqnorm s,h')::hyp) const in
	  hl := h'::!hl;
	  ll := l'::!ll)
	  (List.combine ss rs);
	let p1 = (lookup_tstp "14" (coqnorm h1) hyp) in
	let p2 = (lookup_tstp "15" (coqnorm h2) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_mat,[status(thm)%s],[%s,%s" l (info_str "tab_mat" hyp (List.map (fun h -> [h]) (!hl))) p1 p2;
	flush c;
	Printf.fprintf c "],[%s,%s" p1 p2;
	flush c;
	List.iter (fun l' ->
	  Printf.fprintf c ",%d" l'; flush c
	      ) !ll;
	Printf.fprintf c "%s])).\n" (disch_str2 !hl)
    | Decomposition(h1, ss, rs) ->
	let hl = ref [] in
	let ll = ref [] in
	List.iter (fun (s,r) ->
	  let h' = get_tstp_hyp_name c (coqnorm s) in
	  let l' = ref_tstp1 c r ((coqnorm s,h')::hyp) const in
	  hl := h'::!hl;
	  ll := l'::!ll)
	  (List.combine ss rs);
	let p1 = (lookup_tstp "16" (coqnorm h1) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_dec,[status(thm)%s],[%s" l (info_str "tab_dec" hyp (List.map (fun h -> [h]) (!hl))) p1;
	flush c;
	List.iter (fun l' ->
	    Printf.fprintf c ",%d" l'; flush c
	      ) !ll;
	Printf.fprintf c "%s])).\n" (disch_str2 !hl)
    | Confront(h1,h2,su,tu,sv,tv,r1,r2) ->
	let h3 = get_tstp_hyp_name c (coqnorm su) in
	let h4 = get_tstp_hyp_name c (coqnorm tu) in
	let h5 = get_tstp_hyp_name c (coqnorm sv) in
	let h6 = get_tstp_hyp_name c (coqnorm tv) in
	let l1 = ref_tstp1 c r1 ((coqnorm su,h3)::(coqnorm tu,h4)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm sv,h5)::(coqnorm tv,h6)::hyp) const in
	let p1 = (lookup_tstp "17" (coqnorm h1) hyp) in
	let p2 = (lookup_tstp "18" (coqnorm h2) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_con,[status(thm)%s],[%s,%s,%d,%d%s])).\n" l (info_str "tab_con" hyp [[h3;h4];[h5;h6]]) p1 p2 l1 l2 (disch_str2 [h3;h4;h5;h6])
    | Trans(h1,h2,su,r1) ->
	let h3 = get_tstp_hyp_name c (coqnorm su) in	
	let l1 = ref_tstp1 c r1 ((coqnorm su,h3)::hyp) const in
	let p1 = (lookup_tstp "17" (coqnorm h1) hyp) in
	let p2 = (lookup_tstp "18" (coqnorm h2) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_trans,[status(thm)%s],[%s,%s,%d%s])).\n" l (info_str "tab_trans" hyp [[h3]]) p1 p2 l1 (disch_str2 [h3])
    | NegEqualProp(h,s,t,r1,r2) -> 
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm (neg t)) in
	let h3 = get_tstp_hyp_name c (coqnorm (neg s)) in
	let h4 = get_tstp_hyp_name c (coqnorm t) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm (neg t),h2)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm (neg s),h3)::(coqnorm t,h4)::hyp) const in
	let p1 = (lookup_tstp "21" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_be,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_be" hyp [[h1;h2];[h3;h4]]) p1 l1 l2 (disch_str2 [h1;h2;h3;h4])
    | EqualProp(h,s,t,r1,r2) -> 
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let h3 = get_tstp_hyp_name c (coqnorm (neg s)) in
	let h4 = get_tstp_hyp_name c (coqnorm (neg t)) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm t,h2)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm (neg s),h3)::(coqnorm (neg t),h4)::hyp) const in
	let p1 = (lookup_tstp "22" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_bq,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_bq" hyp [[h1;h2];[h3;h4]]) p1 l1 l2 (disch_str2 [h1;h2;h3;h4])
    | NegAequivalenz(h,s,t,r1,r2) -> 
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm (neg t)) in
	let h3 = get_tstp_hyp_name c (coqnorm (neg s)) in
	let h4 = get_tstp_hyp_name c (coqnorm t) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm (neg t),h2)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm (neg s),h3)::(coqnorm t,h4)::hyp) const in
	let p1 = (lookup_tstp "23" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_negiff,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_negiff" hyp [[h1;h2];[h3;h4]]) p1 l1 l2 (disch_str2 [h1;h2;h3;h4])
    | Aequivalenz(h,s,t,r1,r2) -> 
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm t) in	
	let h3 = get_tstp_hyp_name c (coqnorm (neg s)) in
	let h4 = get_tstp_hyp_name c (coqnorm (neg t)) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::(coqnorm t,h2)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm (neg s),h3)::(coqnorm (neg t),h4)::hyp) const in
	let p1 = (lookup_tstp "24" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_iff,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_iff" hyp [[h1;h2];[h3;h4]]) p1 l1 l2 (disch_str2 [h1;h2;h3;h4])
    | NegEqualFunc(h,s,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let p1 = (lookup_tstp "25" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_fe,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_fe" hyp [[h1]]) p1 l1 (disch_str2 [h1])
    | EqualFunc(h,s,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in	
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let p1 = (lookup_tstp "26" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_fq,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_fq" hyp [[h1]]) p1 l1 (disch_str2 [h1])
    | ChoiceR(eps,pred,s,t,r1,r2) -> 
	let const = add_fresh_const c const pred "" in
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm t) in
	begin
	  match eps with
	  | Choice(a) -> 
	      let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	      let l2 = ref_tstp1 c r2 ((coqnorm t,h2)::hyp) const in
	      Printf.fprintf c "thf(%d,plain,$false,inference(tab_choice,[status(thm)%s],[%d,%d%s])).\n" l (info_str "tab_choice" hyp [[h1];[h2]]) l1 l2 (disch_str2 [h1;h2])
	  | Name(x,Ar(Ar(a,Prop),_)) ->
	      let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	      let l2 = ref_tstp1 c r2 ((coqnorm t,h2)::hyp) const in
	      Printf.fprintf c "thf(%d,plain,$false,inference(tab_choiceop,[status(thm),%s,%s%s],[%d,%d%s])).\n" l x (Choice.get_Choicop_axiom x a hyp) (info_str "tab_choiceop" hyp [[h1];[h2]]) l1 l2 (disch_str2 [h1;h2])
	  | _ -> failwith "eps is not a valid epsilon"
	end
    | Cut(s,r1,r2) -> 
	let const = add_fresh_const c const s "" in
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let h2 = get_tstp_hyp_name c (coqnorm (neg s)) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let l2 = ref_tstp1 c r2 ((coqnorm (neg s),h2)::hyp) const in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_cut,[status(thm)%s],[%d,%d%s])).\n" l (info_str "tab_cut" hyp [[h1];[h2]]) l1 l2 (disch_str2 [h1;h2])
    | DoubleNegation(h,s,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	let p1 = (lookup_tstp "27" (coqnorm h) hyp) in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_dn,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_dn" hyp [[h1]]) p1 l1 (disch_str2 [h1])
    | Rewrite(prefix,pt,pt',r1) ->
	let h =  coqnorm (Ap(prefix,pt)) in
	let h1 = lookup_tstp "28" h hyp in	
	let s =  coqnorm (Ap(prefix,pt')) in 
	let h2 = get_tstp_hyp_name c s in
	let l2 = ref_tstp1 c r1 ((s,h2)::hyp) const in
	let rewout rname =
	  Printf.fprintf c "thf(%d,plain,$false,inference(%s,[status(thm)%s,%s(leibnizp,$thf(" l rname (info_str rname hyp [[h2]]) rname;
	  trm_to_tstp c prefix (Variables.make ());
	  Printf.fprintf c "))],[%s,%d%s])).\n" h1 l2 (disch_str2 [h2])
	in
	begin
	  match pt with
	  | True -> rewout "tab_rew_true"
	  | And -> rewout "tab_rew_and"
	  | Or -> rewout "tab_rew_or"
	  | Iff -> rewout "tab_rew_iff"
	  | Exists(_) -> rewout "tab_rew_ex"
	  | Eq(_) -> rewout "tab_rew_sym"
	  | Lam(_,Lam(_,Ap(DB(1,_),DB(0,_)))) ->  rewout "tab_rew_eta"
	  | Lam(Ar(Prop,Prop),Ap(Ap(Imp,Ap(Ap(Imp,DB(0,Prop)),False)),False)) ->  rewout "tab_rew_dn"
	  | Lam(_,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,(Ap(DB(0,_),DB(2,_)))),(Ap(DB(0,_),DB(1,_)))) ))) )) ->  rewout "tab_rew_leib1"
	  | Lam(_,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,Ap(Ap(Imp,(Ap(DB(0,_),DB(2,_)))),False)),Ap(Ap(Imp,(Ap(DB(0,_),DB(1,_)))),False)) ))) )) ->  rewout "tab_rew_leib2"
	  | Lam(_,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,(Ap(Forall(_),Lam(_,(Ap(Ap(DB(1,_),DB(0,_)),DB(0,_)))))) ),(Ap(Ap(DB(0,_),DB(2,_)),DB(1,_)))) ) )) )) ->  rewout "tab_rew_leib3"
	  | Lam(_,Lam(_, Ap(Forall(_),Lam(_,(Ap(Ap(Imp,(Ap(Ap(Imp,(Ap(Ap(DB(0,_),DB(2,_)),DB(1,_)))),False) )),(Ap(Ap(Imp,(Ap(Forall(_),Lam(_,(Ap(Ap(DB(1,_),DB(0,_)),DB(0,_))))) )),False) )) ) )) )) ->  rewout "tab_rew_leib4"
	  | _ -> failwith("unknown rewrite step found in ref_tstp" ^ (trm_str pt))
	end
    | Delta(h,s,x,r1) ->
	let h1 = (lookup_tstp "29" (coqnorm h) hyp) in
	let h2 = get_tstp_hyp_name c (coqnorm s) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h2)::hyp) const in
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_delta,[status(thm),%s%s],[%s,%d%s])).\n" l x (info_str "tab_delta" hyp [[h2]]) h1 l1 (disch_str2 [h2])
    | KnownResult(s,name,al,r1) ->
	let h1 = get_tstp_hyp_name c (coqnorm s) in
	let l1 = ref_tstp1 c r1 ((coqnorm s,h1)::hyp) const in
	begin
	  Printf.fprintf c "thf(%d,plain,$false,inference(tab_known,[status(thm)%s" l (info_str "tab_known" hyp [[h1]]);
	  flush c;
	  List.iter (fun a ->
	    Printf.fprintf c ",$thf(";
	    flush c;
	    print_stp_tstp c a false;
	    Printf.fprintf c ")";
	    flush c;
	    )
	    al;
	  Printf.fprintf c "],[%s,%d%s])).\n" (tstpizename (itp_thm_str name)) l1 (disch_str2 [h1])
	end
    | NYI(h,s,r1) -> failwith("NYI step found in ref_tstp" )
    | Timeout -> failwith("Timeout step found in ref_tstp" )
  end; flush c; l

    (*** quick check to see if it's worth bothering with making definitions; arbitrarily chosen to start if the conjecture is a propositional combination of at least 7 subformulas. ***)
let rec tstp_enough_subformulas s n =
  if n > 6 then
    (true,7)
  else
    match s with
    | Ap(Ap(binop,s1),s2) when binop = Imp || binop = Or || binop = And || binop = Iff ->
	let (b,k) = tstp_enough_subformulas s1 (n+1) in
	if b then (b,k) else tstp_enough_subformulas s2 k
    | Ap(Neg,s1) ->
	tstp_enough_subformulas s1 (n+1)
    | _ -> (false,n)
    
let rec tstp_define_subformulas c s used =
  begin
    match s with
    | Ap(Ap(binop,s1),s2) when binop = Imp || binop = Or || binop = And || binop = Iff ->
	tstp_define_subformulas c s1 used;
	tstp_define_subformulas c s2 used
    | Ap(Neg,s1) ->
	tstp_define_subformulas c s1 used
    | _ -> ()
  end;
  tstp_print_defprop c used s

 (** Prints refutation r to out_channel c **)
let ref_tstp c r =
  List.iter
    (fun (s,pt) ->
      let m = coqnorm pt in
      Hashtbl.add tstp_axioms m (s,m,c,ref true))
    (!coqsig_hyp);
  match !conjecture with
  | Some(con,_) ->
      let ccon = coqnorm con in (*conjecture*)
      begin
	if let (b,_) = tstp_enough_subformulas ccon 0 in b then (* Chad, May 3 2016: Some problems have large conjectures that cause significant time to be wasted printing the same subformulas multiple times; in these cases create definitions sPn. *)
	  let usednames = ref [] in
	  Hashtbl.iter (fun k _ -> usednames := k::!usednames) coq_used_names;
	  Hashtbl.iter (fun _ (x,_,_,_) -> usednames := x::!usednames) tstp_axioms;
	  tstp_define_subformulas c ccon !usednames;
      end;
      let ncon = neg ccon in (*negated conjecture*)
      let h1 = get_thyp_name() in (*hypothesis name*)
      Printf.fprintf c "thf(%s,conjecture," (!conjecturename);
      flush c;
      trm_to_tstp c ccon (Variables.make ());
      flush c;
      Printf.fprintf c ").\n";
      flush c;
      Printf.fprintf c "thf(%s,negated_conjecture," h1;
      flush c;
      trm_to_tstp_rembvar h1 c ncon (Variables.make ());
      flush c;
      Printf.fprintf c ",inference(assume_negation,[status(cth)],[%s])).\n" (!conjecturename);
      flush c;
      let ll = ref_tstp1 c r [(ncon,h1)] (!coqsig_const) in
      Printf.fprintf c "thf(0,theorem,";
      flush c;
      trm_to_tstp c ccon (Variables.make ()); (*print the conjecture again*)
      flush c;
      Printf.fprintf c ",inference(contra,[status(thm),contra(discharge,[%s])],[%d,%s])).\n" h1 ll h1;
      flush c;
  | None ->
      ignore (ref_tstp1 c r [] (!coqsig_const))



let tstp_axiom_variants : (trm,string) Hashtbl.t = Hashtbl.create 100;;

	(*** different version of lookup_tstp for use with refut_tstp below ***)
let lookup_tstp_2016 w s hyp =
  try
    List.assoc s hyp
  with
  | Not_found ->
      try
	Hashtbl.find tstp_axiom_variants s
      with Not_found ->
	begin
	  try
	    let (x,m,c,f) = Hashtbl.find tstp_axioms s in
	    let rl = ref x in
	    if (!f) then
	      begin
		f := false;
		Printf.fprintf c "thf(%s,axiom," x;
		trm_to_tstp_rembvar x c m (Variables.make ());
		Printf.fprintf c ").\n";
		if not (m = s) then
		  begin
		    let l = !tstpline in
		    let ll = string_of_int l in
		    incr tstpline;
		    Printf.fprintf c "thf(%d,plain," l;
		    trm_to_tstp_rembvar ll c s (Variables.make ());
		    Printf.fprintf c ",inference(preprocess,[status(thm)],[%s]).\n" x;
		    rl := ll
		  end;
		flush c
	      end;
	    Hashtbl.add tstp_axiom_variants s !rl;
	    !rl
	  with
	  | Not_found ->
	      Printf.printf "%s: Could not find hyp name\ns = %s\nhyp:\n" w (trm_str s);
	      List.iter (fun (m,h) -> Printf.printf "%s: %s\n" h (trm_str m)) hyp;
	      failwith ("Could not find hyp name")
	end

let lookup_tstp_assumption_lit_2016 w c l hyp =
  try
    let (m1,m2) = Hashtbl.find assumption_lit l in
    begin
      try
	List.assoc m1 hyp
      with Not_found ->
	try
	  Hashtbl.find tstp_axiom_variants m1
	with Not_found ->
	  let p = lookup_tstp_2016 w m2 hyp in
	  if m1 = m2 then
	    p
	  else
	    begin
	      let l = !tstpline in
	      let ll = string_of_int l in
	      incr tstpline;
	      Printf.fprintf c "thf(%d,plain," l;
	      trm_to_tstp_rembvar ll c m1 (Variables.make ());
	      Printf.fprintf c ",inference(normalize,[status(thm)],[%s]).\n" p;
	      Hashtbl.add tstp_axiom_variants m1 ll;
	      ll
	    end
    end
  with Not_found ->
    Printf.printf "Could not find assumption lit %d\n" l;
    List.iter (fun (m,h) -> Printf.printf "%s: %s\n" h (trm_str m)) hyp;
    failwith ("Could not find assumption info for proof reconstruction")

let tstp_fodef_2016 : (int,string) Hashtbl.t = Hashtbl.create 100;;

let lookup_tstp_fodef_2016 c a =
  try
    Hashtbl.find tstp_fodef_2016 a
  with Not_found ->
    let ll =
      if a > 0 then
	let m = Atom.atom_to_trm a in
	Hashtbl.find tstp_defprops m
      else
	let m = Atom.atom_to_trm (-a) in
	Hashtbl.find tstp_defprops m
    in
    Hashtbl.add tstp_fodef_2016 a ll;
    ll

let refut_tstp_clause c cr eigenchoicetps hyp clause =
  let ri = cr clause in
  let sp lit =
    let a = if lit > 0 then lit else -lit in
    let s = Atom.atom_to_trm a in
    try
      let p = Hashtbl.find tstp_defprops s in
      if lit > 0 then
        p
      else
        "~(" ^ p ^ ")"
    with Not_found ->
      raise (Failure("Could not find propositional literal corresponding to " ^ (string_of_int lit)))
  in
  let cs =
    match clause with
    | [] -> "$false" (*** should not happen ***)
    | lit::clauser ->
        List.fold_left (fun x lit -> "(" ^ x ^ " | " ^ (sp lit) ^ ")") (sp lit) clauser
  in
  let ordinarycase justif =
    let l = !tstpline in
    incr tstpline;
    Printf.fprintf c "thf(%d,plain,%s,inference(%s,[status(thm)],[])).\n" l cs justif;
    string_of_int l
  in
  match ri with
  | DeltaRule -> ordinarycase "delta_rule"
  | NegPropRule(m) -> ordinarycase "prop_rule"
  | PosPropRule(m) -> ordinarycase "prop_rule"
  | MatingRule(plit,nlit) -> ordinarycase "mating_rule"
  | ConfrontationRule(plit,nlit) -> ordinarycase "confrontation_rule"
  | FreshRule(a,m,x) ->
      begin
        try
          let (epsname,heps) = List.assoc a eigenchoicetps in
          let tstp_name = tstpizename x in
          let deflinename = "eigendef_" ^ tstp_name in
          let l = !tstpline in
          incr tstpline;
          Printf.fprintf c "thf(%d,plain,%s,inference(eigen_choice_rule,[status(thm),assumptions([%s])],[%s,%s])).\n" l cs heps heps deflinename;
          string_of_int l
        with Not_found ->
          ordinarycase "eigen_choice_rule" (*** this probably shouldn't happen, but if it does then just give what information can be given ***)
      end
  | ChoiceRule(Choice(_),pred) -> ordinarycase "choice_rule"
  | InstRule(a,m,n) -> ordinarycase "all_rule" (*** this isn't really enough information; the instantiation "n" should be given, but I don't know a convenient TSTP compliant way to give it here ***)
  | Known(i,x,al) -> ordinarycase (itp_thm_str x) (*** just use x as the name of the rule; should also give the type instantiations "al", but I don't know a convenient TSTP compliant way to give it here ***)
  | ChoiceRule(Name(x,Ar(Ar(a,_),_)),pred) ->
      begin
        try
          let (a,m,mb) = Hashtbl.find Choiceop.choiceopnames x in
          begin
            try
              let p1 = List.assoc m hyp in
              let l = !tstpline in
              incr tstpline;
              Printf.fprintf c "thf(%d,plain,%s,inference(choice_rule,[status(thm)],[%s])).\n" l cs p1;
              string_of_int l
            with Not_found ->
              try
                let p1 = Hashtbl.find tstp_axiom_variants m in
                let l = !tstpline in
                incr tstpline;
                Printf.fprintf c "thf(%d,plain,%s,inference(choice_rule,[status(thm)],[%s])).\n" l cs p1;
                string_of_int l
              with Not_found ->
                let p1 = lookup_tstp_2016 "98" mb hyp in
                if m = mb then
                  let l = !tstpline in
                  incr tstpline;
                  Printf.fprintf c "thf(%d,plain,%s,inference(choice_rule,[status(thm)],[%s])).\n" l cs p1;
                  string_of_int l
                else
                  begin
                    let l = !tstpline in
                    incr tstpline;
                    let p2 = string_of_int l in
                    Printf.fprintf c "thf(%d,plain," l;
                    trm_to_tstp_rembvar p2 c m (Variables.make ());
                    Printf.fprintf c ",inference(preprocess,[status(thm)],[%s]).\n" p2;
                    Hashtbl.add tstp_axiom_variants m p2;
                    let l = !tstpline in
                    incr tstpline;
                    Printf.fprintf c "thf(%d,plain,%s,inference(choice_rule,[status(thm)],[%s])).\n" l cs p2;
                    string_of_int l
                  end
          end
        with Not_found ->
          raise (Failure (x ^ " is used as a choice operator, but no corresponding proposition was recorded"))
      end
  | ChoiceRule(_,_) ->
      raise (Failure ("Bad term used as choice operator"))

let refut_tstp c r =
  let prehyp = ref [] in
  let eigenhyp = ref [] in
  Hashtbl.iter
    (fun mn (s,m) -> Hashtbl.add tstp_axioms mn (s,m,c,ref true))
    name_hyp_inv;
  let {basetypes;consttypes;eigendefs;propreferenced} = refutsiginfo_init () in
  refutsiginfo_of_refut {basetypes;consttypes;eigendefs;propreferenced} r;
  Hashtbl.remove basetypes "$i"; (* TPTP already declares this *)
  Hashtbl.iter
    (fun basetype_name _ ->
      Hashtbl.add coq_used_names basetype_name (); (* Chad, May 3 2016: mark this as used to avoid conflicts if new names are created *)
      output_string c
	("thf(ty_" ^ basetype_name ^ ", type, " ^ basetype_name ^ " : $tType).\n"))
    basetypes;
  Hashtbl.iter
    (fun name stp ->
      let tstp_name = tstpizename name in
      Hashtbl.add coq_used_names tstp_name (); (* Chad, May 3 2016: mark this as used to avoid conflicts if new names are created *)
      output_string c ("thf(ty_" ^ tstp_name ^ ", type, " ^ tstp_name ^ " : ");
      print_stp_tstp c stp true;
      output_string c (").\n"))
    consttypes;
  let eigenchoicetps = ref [] in
  Hashtbl.iter
    (fun name def ->
      let tstp_name = tstpizename name in
      match def with
      | Ap(Choice(a),m) ->
	  let eigendef e =
	    tstp_print_eigendef c tstp_name (Ap(Name(e,Ar(Ar(a,Prop),a)),m))
	  in
	  begin
	    try
	      let (e,_) = List.assoc a !eigenchoicetps in
	      eigendef e
	    with Not_found ->
	      begin
		let e = "eps__" ^ (string_of_int (List.length !eigenchoicetps)) in
		let s = forall (Ar(a,Prop)) (forall a (imp (Ap(DB(1,Ar(a,Prop)),DB(0,a))) (Ap(DB(1,Ar(a,Prop)),Ap(Name(e,Ar(Ar(a,Prop),a)),DB(1,Ar(a,Prop))))))) in
		let h = get_thyp_name() in
		eigenchoicetps := (a,(e,h))::!eigenchoicetps;
		output_string c ("thf(" ^ h ^ ", assumption, ");
		trm_to_tstp c s (Variables.make());
		output_string c (",introduced(assumption,[])).\n");
		eigenhyp := (tstp_name,h)::!eigenhyp;
		prehyp := (s,h)::!prehyp;
		eigendef e
	      end
	  end
      | _ ->
	  tstp_print_eigendef c tstp_name def
	    )
    eigendefs;
  let usednames = ref [] in
  Hashtbl.iter (fun k _ -> usednames := k::!usednames) coq_used_names;
  Hashtbl.iter (fun _ (x,_,_,_) -> usednames := x::!usednames) tstp_axioms;
  Hashtbl.iter
    (fun a _ ->
      let s = Atom.atom_to_trm a in
      tstp_print_defprop c !usednames s
      )
    propreferenced;
  List.iter
    (fun z ->
      match z with
      | ProbDef(x,tp,tm,al,w) ->
	  Printf.fprintf c "thf(def_%s,definition,(%s = " x x;
	  trm_to_tstp_rembvar x c tm (Variables.make());
	  Printf.fprintf c ")).\n";
      | _ -> ())
    !probsig;
  let rec discharge_eigenhyps p ehyp hyp =
    match (ehyp,hyp) with
    | (((x,h)::ehypr),(_::hypr)) ->
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(eigenvar_choice,[status(thm)%s],[%d%s])).\n" l (info_str "eigenvar_choice" hypr [[h]]) p (disch_str2 [h]);
	discharge_eigenhyps l ehypr hypr
    | _ -> ()
  in
  let rec discharge_eigenhyps_c p ehyp nc hyp =
    match (ehyp,hyp) with
    | (((x,h)::ehypr),(_::hypr)) ->
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(eigenvar_choice,[status(thm)%s],[%d%s])).\n" l (info_str "eigenvar_choice" (nc::hypr) [[h]]) p (disch_str2 [h]);
	discharge_eigenhyps_c l ehypr nc hypr
    | _ -> ()
  in
  let rec refut_tstp_r hyp const r =
(**    Printf.printf "** refut_tstp_r\nr:\n"; print_refut r; Printf.printf "hyp:\n"; List.iter (fun (m,h) -> Printf.printf "%s: %s\n" h (trm_str m)) hyp; **)
    match r with
    | Eprover(pcore,cr,fcore,elines) ->
	begin
	  let pl =
	    List.map
	      (fun clause ->
		begin
		  try refut_tstp_clause c cr !eigenchoicetps hyp clause
		  with Not_found ->
		    match clause with
		    | [lit] ->
			let p = lookup_tstp_assumption_lit_2016 "187" c lit hyp in
			p
		    | _ ->
			raise (Failure "no rule for non-unit clause")
		end)
	      pcore
	  in
	  let fl =
	    List.map
	      (fun a -> lookup_tstp_fodef_2016 c a)
	      fcore
	  in
	  let lasteline = ref "" in
	  List.iter (fun l -> Printf.fprintf c "%s\n" l; lasteline := l) elines;
	  let pfls =
	    String.concat "," (
	      try
		Eprover.folabel !lasteline :: pl @ fl
	      with Not_found ->
		(pl @ fl)
	    )
	  in
	  let l = !tstpline in
	  incr tstpline;
	  Printf.fprintf c "thf(%d,plain,$false,inference(eprover,[status(thm)%s],[%s])).\n" l (info_str "fo_unsat" hyp []) pfls;
	  l
	end
    | SearchR(cl, cr) ->
	let pl =
	  List.map
	    (fun clause ->
	      begin
		try refut_tstp_clause c cr !eigenchoicetps hyp clause
		with Not_found ->
		  match clause with
		  | [lit] ->
		      let p = lookup_tstp_assumption_lit_2016 "1" c lit hyp in
		      p
		  | _ ->
		      raise (Failure "no rule for non-unit clause")
	      end)
	    cl
	in
	let l = !tstpline in
	incr tstpline;
	let pls =
          String.concat "," pl
	in
	Printf.fprintf c "thf(%d,plain,$false,inference(prop_unsat,[status(thm)%s],[%s])).\n" l (info_str "prop_unsat" hyp []) pls;
	l
    | AssumptionConflictR(lit) ->
	let a = if lit > 0 then lit else -lit in
	let p1 = lookup_tstp_assumption_lit_2016 "2" c a hyp in
	let p2 = lookup_tstp_assumption_lit_2016 "3" c (-a) hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_conflict,[status(thm)%s],[%s,%s])).\n" l (info_str "tab_conflict" hyp []) p1 p2;
	l
    | FalseR ->
	let p1 = lookup_tstp_2016 "4" False hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_false,[status(thm)%s],[%s])).\n" l (info_str "tab_false" hyp []) p1;
	l	
    | NegReflR(t) ->
	let p1 = lookup_tstp_2016 "5" t hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_negrefl,[status(thm)%s],[%s])).\n" l (info_str "tab_negrefl" hyp []) p1;
	l	
    | ImpR(t1,t2,r1,r2) ->
	let s1 = coqnorm t1 in
	let ns1 = normneg s1 in
	let s2 = coqnorm t2 in
	let h1 = get_tstp_hyp_name c ns1 in
	let h2 = get_tstp_hyp_name c s2 in
	let l1 = refut_tstp_r ((ns1,h1)::hyp) const r1 in
	let l2 = refut_tstp_r ((s2,h2)::hyp) const r2 in
	let p1 = lookup_tstp_2016 "6" (Ap(Ap(Imp,s1),s2)) hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_imp,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_imp" hyp [[h1];[h2]]) p1 l1 l2 (disch_str2 [h1;h2]);
	l
    | EqOR(t1,t2,r1,r2) ->
	let s1 = coqnorm t1 in
	let ns1 = normneg s1 in
	let s2 = coqnorm t2 in
	let ns2 = normneg s2 in
	let h1 = get_tstp_hyp_name c s1 in
	let h2 = get_tstp_hyp_name c s2 in
	let h3 = get_tstp_hyp_name c ns1 in
	let h4 = get_tstp_hyp_name c ns2 in
	let l1 = refut_tstp_r ((s1,h1)::(s2,h2)::hyp) const r1 in
	let l2 = refut_tstp_r ((ns1,h3)::(ns2,h4)::hyp) const r2 in
	let p1 = lookup_tstp_2016 "23" (Ap(Ap(Eq(Prop),s1),s2)) hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_bq,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_bq" hyp [[h1;h2];[h3;h4]]) p1 l1 l2 (disch_str2 [h1;h2;h3;h4]);
	l
    | NegEqOR(t1,t2,r1,r2) ->
	let s1 = coqnorm t1 in
	let ns1 = normneg s1 in
	let s2 = coqnorm t2 in
	let ns2 = normneg s2 in
	let h1 = get_tstp_hyp_name c s1 in
	let h2 = get_tstp_hyp_name c s2 in
	let h3 = get_tstp_hyp_name c ns1 in
	let h4 = get_tstp_hyp_name c ns2 in
	let l1 = refut_tstp_r ((s1,h1)::(ns2,h2)::hyp) const r1 in
	let l2 = refut_tstp_r ((ns1,h3)::(s2,h4)::hyp) const r2 in
	let p1 = lookup_tstp_2016 "22" (neg(Ap(Ap(Eq(Prop),s1),s2))) hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_be,[status(thm)%s],[%s,%d,%d%s])).\n" l (info_str "tab_be" hyp [[h1;h2];[h3;h4]]) p1 l1 l2 (disch_str2 [h1;h2;h3;h4]);
	l
    | NegAllR(a,t,x,r) ->
	let s1 = coqnorm (norm name_def (ap(t,Name(x,a)))) in
	let ns1 = normneg s1 in
	let h1 = get_tstp_hyp_name c ns1 in
	let l1 = refut_tstp_r ((ns1,h1)::hyp) ((x,a)::const) r in
	let p1 = lookup_tstp_2016 "11" (neg(Ap(Forall(a),t))) hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_negall,[status(thm)%s,tab_negall(eigenvar,%s)],[%s,%d%s])).\n" l (info_str "tab_negall" hyp [[h1]]) (tstpizename x) p1 l1 (disch_str2 [h1]);
	l
    | NegEqFR(a,b,t1,t2,r) ->
	let t = coqnorm (forall a (eq b (ap(t1,DB(0,a))) (ap(t2,DB(0,a))))) in
	let nt = neg t in
	let s = coqnorm (eq (Ar(a,b)) t1 t2) in
	let ns = neg s in
	let h1 = get_tstp_hyp_name c nt in
	let l1 = refut_tstp_r ((nt,h1)::hyp) const r in
	let p1 = lookup_tstp_2016 "25" ns hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_fe,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_fe" hyp [[h1]]) p1 l1 (disch_str2 [h1]);
	l
    | NegImpR(t1,t2,r) ->
	let s1 = coqnorm t1 in
	let s2 = coqnorm t2 in
	let ns2 = normneg s2 in
	let h1 = get_tstp_hyp_name c s1 in
	let h2 = get_tstp_hyp_name c ns2 in
	let l1 = refut_tstp_r ((s1,h1)::(ns2,h2)::hyp) const r in
	let p1 = lookup_tstp_2016 "7" (normneg(Ap(Ap(Imp,s1),s2))) hyp in
	let l = !tstpline in
	incr tstpline;
	Printf.fprintf c "thf(%d,plain,$false,inference(tab_negimp,[status(thm)%s],[%s,%d%s])).\n" l (info_str "tab_negimp" hyp [[h1;h2]]) p1 l1 (disch_str2 [h1;h2]);
	l
  in
  match !conjecture with
  | Some(con,nncon) ->
      let ccon = coqnorm con in (*conjecture*)
      let ncon = coqnorm nncon in (*negated conjecture*)
      let h1 = get_thyp_name() in (*hypothesis name*)
      Printf.fprintf c "thf(%s,conjecture," (!conjecturename);
      flush c;
      trm_to_tstp c ccon (Variables.make ());
      flush c;
      Printf.fprintf c ").\n";
      flush c;
      Printf.fprintf c "thf(%s,negated_conjecture," h1;
      flush c;
      trm_to_tstp_rembvar h1 c ncon (Variables.make ());
      flush c;
      Printf.fprintf c ",inference(assume_negation,[status(cth)],[%s])).\n" (!conjecturename);
      flush c;
      let l = refut_tstp_r ((ncon,h1)::!prehyp) [] r in
      discharge_eigenhyps_c l !eigenhyp (ncon,h1) !prehyp;
      Printf.fprintf c "thf(0,theorem,";
      flush c;
      trm_to_tstp c ccon (Variables.make ()); (*print the conjecture again*)
      flush c;
      Printf.fprintf c ",inference(contra,[status(thm),contra(discharge,[%s])],[%d,%s])).\n" h1 l h1;
      flush c;
  | None ->
      let l = refut_tstp_r !prehyp [] r in
      discharge_eigenhyps l !eigenhyp !prehyp
