(* Stateful term printing *)

let coq_used_names : (string,unit) Hashtbl.t = Hashtbl.create 100
let coq_names : (string,string) Hashtbl.t = Hashtbl.create 100
let coq_hyp_names : (string,string) Hashtbl.t = Hashtbl.create 100

(** Translate de Bruijn indices to variables**)
module Variables = struct
  (** next variable counter and list of used variable names**)
  type t = int * (string list)
  let make () = (1,[])
  (** Input: Variables (n,v)
    Output: Variables (m,v') with m>n and a new variable name x in v'  **)
  let rec push (n,v) =
    let x = "X" ^ (string_of_int n) in
    let n = n+1 in
    if (Hashtbl.mem coq_used_names x)
      then push (n,v)
      else (n,x::v)
  let top (_,v) = List.hd v
  let get i (_,v) = List.nth v i
end

let coq_keywords =
[ "as"
; "at"
; "cofix"
; "else"
; "end"
; "exists2"
; "fix"
; "for"
; "forall"
; "fun"
; "if"
; "IF"
; "in"
; "let"
; "match"
; "mod"
; "Prop"
; "return"
; "Set"
; "then"
; "Type"
; "using"
; "where"
; "with"
(*** Avoid certain names used by stt in Coq ***)
; "SType"
; "Sar"
; "o"
; "prop"
; "Sepsilon"
; "forall"
; "exists"
; "False"
; "True"
; "not"
; "Snot"
; "and"
; "Sand"
; "or"
; "Sor"
; "iff"
; "Siff"
; "ex"
; "SSigma"
; "SPi"
; "eq"
; "Seq"
; "I"
; "FalseE"
; "conj"
; "proj1"
; "proj2"
; "and_ind"
; "or_introl"
; "or_intror"
; "or_ind"
; "ex_intro"
; "ex_ind"
; "refl_equal"
; "eq_ind"
; "SinhE"
; "classic"
; "NNPP"
; "prop_ext"
; "functional_extensionality"
; "Sepsilon_spec"
(*** Other names to avoid ***)
; "claim"
(*FIXME add isar keywords*)
; "thm"
; "lemma"
; "case"
]

let termP_init () =
  List.iter (fun w -> Hashtbl.add coq_used_names w ()) coq_keywords

let tstpizename x =
  if (String.length x > 0) then
    if (String.get x 0 = '_') then
      "eigen" ^ x
    else if (String.get x 0 = '@') then
      String.sub x 1 ((String.length x) - 1)
    else
      if ((String.length x > 5) && (String.sub x 0 5 = "eigen")) then
        "not" ^ x
      else
        x
  else x
