exception Timeout

(** [set_alarm f s] causes a function [f] to be called after [s] seconds,
    see http://alan.petitepomme.net/cwn/2006.01.03.html#2 *)
let set_alarm f s =
  ignore (Sys.signal Sys.sigalrm (Sys.Signal_handle f));
  ignore (Unix.setitimer Unix.ITIMER_REAL { Unix.it_interval = 0.0; Unix.it_value = s })

(** [remaining_time ()] returns the seconds remaining until
    last previously set alarm will go off *)
let remaining_time () =
  let a = Unix.getitimer Unix.ITIMER_REAL in
  a.Unix.it_value

(** [set_timeout s] causes a Timeout to be raised after at most [s] seconds *)
let set_timeout s =
  if s > 0.0 then set_alarm (fun signo -> raise Timeout) s
  else raise Timeout

(** [set_sub_timeout s] raises a timeout after at most [s] seconds as well as
    a timeout that was previously set.
    Returns a function to cancel the new and to restore the original timeout. *)
let set_sub_timeout s =
  if s > 0.0 then
    (*                               orig.
       now                           alarm
        |------------------------------|
    
        \--------|--------/\-----|-----/
                 s              z_2
              (= s_2)
    
        \--------------|---------------/
                       z
    *)
    let z = remaining_time () in
    let s2 = min z s in
    let z2 = z -. s2 in
    set_alarm (fun signo -> set_timeout z2; raise Timeout) s2;
    (fun () -> set_timeout (z2 +. remaining_time ()))
  else
    raise Timeout;
